include 'BIOS.inc'
include 'Vram.inc'
include 'Chars.inc'

ext Chars, SpritePattern, BackgroundChars, FortPattern
ext Ground_, CharOffset_, DotOffset_
ext FortDotOffset_, FortCharOffset_

StageWidth equ 20
StageHeight equ 18-2-2
GroundWidth equ StageWidth+2
Top equ 7
Bottom equ StageHeight-2
GroundHeight equ Bottom-Top
BackgroundCharCount equ (Char_Background_End - Char_Background) / 2
FortCharCount equ (Char_End - Char_Fort) / 2


; void InitMonoChar(word vram, ptr<byte> p, byte count, byte color);
dseg
InitMonoChar_@Param3: public InitMonoChar_@Param3
    defb 0
cseg
InitMonoChar_: public InitMonoChar_
    push af | push hl | push de | push bc
        ld a,l | out (VdpPort+1),a
        ld a,h | or 40h | out (VdpPort+1),a
        ex de,hl
        do
            push bc
                ld c,CharHeight
                do
                    ld e,(hl) | inc hl
                    ld a,(InitMonoChar_@Param3)
                    ld d,a
                    ld b,4
                    do
                        bit 0,d
                        if nz
                            ld a,e
                        else
                            xor a
                        endif
                        out (VdpPort),a
                        srl d
                    dwnz
                    dec c
                while nz | wend
            pop bc
            dec c
        while nz | wend
    pop bc | pop de | pop hl | pop af
ret


; void InitColorChar(word vram, ptr<byte> p, byte count);
cseg
InitColorChar_: public InitColorChar_
    push af | push hl | push de | push bc
        ld a,l | out (VdpPort+1),a
        ld a,h | or 40h | out (VdpPort+1),a
        ex de,hl
        do
            ld b,CharHeight*4
            do
                ld a,(hl) | inc hl
                out (VdpPort),a
            dwnz
            dec c
        while nz | wend
    pop bc | pop de | pop hl | pop af
ret


; void ClearScreen();
cseg
ClearScreen_:
    public ClearScreen_
    push af | push hl | push bc
        ld hl,Vram_PatternName
        ld bc,VramHeight*VramRowSize
        xor a
        call FILVRM

        ld hl,Vram_SpriteAttribute
        ld bc,64
        ld a,0d0h
        call FILVRM
    pop bc | pop hl | pop af
ret


; word VramAddress(byte x, byte y);
VramAddress_: public VramAddress_
    push de
        ld l,e
        ld h,0
        add hl,hl ;*2
        add hl,hl ;*4
        add hl,hl ;*8
        add hl,hl ;*16
        add hl,hl ;*32
        add hl,hl ;*64
        add a,a
        ld e,a
        ld d,0
        add hl,de
        ld de,Vram
        add hl,de
    pop de
ret


; void Put(word vram, byte c);
cseg
Put_: public Put_
    push af
        ld a,l | out (VdpPort+1),a
        ld a,h | or 40h | out (VdpPort+1),a
        ld a,e
        out (VdpPort),a
        nop | nop | nop
        xor a
        out (VdpPort),a
    pop af
    inc hl
    inc hl
ret


; word PrintC(word vram, byte c);
cseg
PrintC_: public PrintC_
    push af
        ld a,l | out (VdpPort+1),a
        ld a,h | or 40h | out (VdpPort+1),a
        ld a,e
        sub ' '
        out (VdpPort),a
        nop | nop | nop
        xor a
        out (VdpPort),a
    pop af
    inc hl
    inc hl
ret


; void GroundToVram();
cseg
GroundToVram_: public GroundToVram_
    push af | push hl | push de | push bc
        ld hl,Vram+VramRowSize*2+4000h
        ld de,VramRowSize
        ld c,Top
        do
            di
                ld a,l | out (VdpPort+1),a
                ld a,h | out (VdpPort+1),a
            ei
            ld b,StageWidth*VramStep
            do
                xor a
                out (VdpPort),a
                nop | nop | nop
            dwnz
            add hl,de
            dec c
        while nz | wend

        ld a,(CharOffset_)
        exx
            ld c,a
            ld hl,Ground_
            ; ld de,GroundWidth-StageWidth
        exx
        ld c,GroundHeight
        do
            di
                ld a,l | out (VdpPort+1),a
                ld a,h | out (VdpPort+1),a
            ei
            ld b,StageWidth
            do
                exx
                    ld a,(hl) | inc hl
                exx
                cp Char_Background
                if nc
                    exx
                        add a,c
                    exx
                endif
                out (VdpPort),a
                xor a
                nop | nop | nop
                out (VdpPort),a
            dwnz
            add hl,de
            exx
                inc hl | inc hl
                ; add hl,de
            exx
            dec c
        while nz | wend

        ld c,StageHeight-Bottom
        do
            di
                ld a,l | out (VdpPort+1),a
                ld a,h | out (VdpPort+1),a
            ei
            ld b,StageWidth
            do
                ld a,Char_Background
                out (VdpPort),a
                xor a
                nop | nop | nop
                out (VdpPort),a
            dwnz
            add hl,de
            dec c
        while nz | wend
    pop bc | pop de | pop hl | pop af
ret


; void RollChar();
cseg
RollChar_: public RollChar_
    push af | push hl | push de | push bc
        ld a,(CharOffset_)
        add a,Char_Background
        ld l,a
        ld h,0
        add hl,hl ;*2
        add hl,hl ;*4
        add hl,hl ;*8
        add hl,hl ;*16
        add hl,hl ;*32
        set 6,h
        di
            ld a,l | out (VdpPort+1),a
            ld a,h | out (VdpPort+1),a
        ei
        exx
            ld a,(DotOffset_)
            ld c,a
            ld hl,BackgroundChars
        exx
        ld b,BackgroundCharCount*CharHeight
        do
            xor a
            out (VdpPort),a
            nop | nop | nop | nop
            out (VdpPort),a
            nop | nop | nop | nop
            out (VdpPort),a
            exx
                ld e,(hl) | inc hl
                ld d,(hl) | inc hl
                ld a,c
                or a
                do
                while nz
                    sla d | rl e
                    dec a
                wend
                ld a,e
            exx
            out (VdpPort),a
            nop | nop
        dwnz
    pop bc | pop de | pop hl | pop af
ret


; void SetFortChars();
cseg
SetFortChars_: public SetFortChars_
    push af | push hl | push de | push bc
        ld a,(FortCharOffset_)
        add a,Char_Fort
        ld l,a
        ld h,0
        add hl,hl ;*2
        add hl,hl ;*4
        add hl,hl ;*8
        add hl,hl ;*16
        add hl,hl ;*32
        set 6,h
        ld c,VdpPort+1
        di
            out (c),l
            out (c),h
        ei
        exx
            ld a,(FortDotOffset_)
            ld c,a
            ld hl,FortPattern+PatternSize
        exx
        ld hl,FortPattern
        ld bc,FortCharCount*PatternSize
        do
            ld a,(hl) | inc hl
            exx
                ld e,a
                ld d,(hl) | inc hl
                ld a,c
                or a
                do
                while nz
                    sla d | rl e
                    dec a
                wend
                ld a,e
            exx
            out (VdpPort),a
            dec bc
            ld a,c
            or b
        while nz | wend
    pop bc | pop de | pop hl | pop af
ret
